from . import query


class Page:
    @staticmethod
    def from_raw(raw):
        return Page(raw["data"], raw.get("before"), raw.get("after"))

    def __init__(self, data, before=None, after=None):
        self.data = data
        self.before = before
        self.after = after

    def map_data(self, func):
        return Page([func(x) for x in self.data], self.before, self.after)

    def __repr__(self):
        return "Page(data=%s, before=%s, after=%s)" % (
            self.data,
            self.before,
            self.after,
        )

    def __eq__(self, other):
        return (
            isinstance(other, Page)
            and self.data == other.data
            and self.before == other.before
            and self.after == other.after
        )

    @staticmethod
    def set_iterator(client, set_query, map_lambda=None, mapper=None, page_size=None):
        def get_page(**kwargs):
            queried = query.paginate(set_query, **kwargs)
            if map_lambda is not None:
                queried = query.map_(map_lambda, queried)
            return Page.from_raw(client.query(queried))

        page = get_page(size=page_size)
        for val in page.data:
            yield val if mapper is None else mapper(val)
        next_cursor = "after" if page.after is not None else "before"
        while getattr(page, next_cursor) is not None:
            page = get_page(
                **{"size": page_size, next_cursor: getattr(page, next_cursor)}
            )
            for val in page.data:
                yield val if mapper is None else mapper(val)



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Type,
    TypeVar,
    Union,
    cast,
)
from uuid import UUID
from aiohttp.web import Response
from pydantic import BaseConfig  # pylint: disable=no-name-in-module
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from pydantic import Extra  # pylint: disable=no-name-in-module
from pydantic import Field
from typing_extensions import ParamSpec
from .faunadb.objects import FaunaTime, Ref
from jinja2 import Template, Environment, FileSystemLoader

Vector = List[float]

MetaData = Dict[str, str]

Undefined = Union[None, type(...), type(Ellipsis)]
NoArgAnyCallable = Callable[[], Any]

T = TypeVar("T")
P = ParamSpec("P")
B = TypeVar("B", bound=BaseModel)


class Document(BaseModel):
    """Document
    Base document for all FaunaDB documents
    """

    class Config(BaseConfig):
        """Configuration defaults for all documents"""

        extra = Extra.allow
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        json_encoders = {
            datetime: lambda v: v.astimezone().isoformat(),
            date: lambda v: v.isoformat(),
            time: lambda v: v.isoformat(),
            timedelta: lambda v: v.total_seconds(),
            UUID: lambda v: str(v),
            Decimal: lambda v: float(v),
            Ref: lambda v: v.id(),
            FaunaTime: lambda v: v.value,
            Enum: lambda v: v.value,
        }

    class Metadata:
        subclasses: List[Type[Document]] = []

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls.__doc__ is None:
            cls.__doc__ = f"```json\n{cls.schema_json(indent=2)}\n```"
        cls.Metadata.subclasses.append(cls)


D = TypeVar("D", bound=Document)


class LazyProxy(Generic[T], ABC):
    def __init__(self) -> None:
        self.__proxied: T | None = None

    def __getattr__(self, attr: str) -> object:
        return getattr(self.__get_proxied__(), attr)

    def __repr__(self) -> str:
        return repr(self.__get_proxied__())

    def __str__(self) -> str:
        return str(self.__get_proxied__())

    def __dir__(self) -> Iterable[str]:
        return self.__get_proxied__().__dir__()

    def __get_proxied__(self) -> T:
        proxied = self.__proxied
        if proxied is not None:
            return proxied

        self.__proxied = proxied = self.__load__()
        return proxied

    def __set_proxied__(self, value: T) -> None:
        self.__proxied = value

    def __as_proxied__(self) -> T:
        """Helper method that returns the current proxy, typed as the loaded object"""
        return cast(T, self)

    @abstractmethod
    def __load__(self) -> T:
        ...


class Component(BaseModel):
    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls.__doc__ is None:
            raise ValueError("Component must have a docstring")

    def render(self) -> Response:
        template = Template(self.__doc__)
        text = template.render(**self.dict())
        return Response(text=text, content_type="text/html")

    def __call__(self) -> Response:
        return self.render()



import base64
from base64 import urlsafe_b64decode, urlsafe_b64encode
from datetime import date, datetime
from enum import Enum
from json import JSONEncoder, dumps, loads
from time import time
from typing import Any, Dict, List, Literal, TypeVar
from uuid import UUID

from iso8601 import parse_date
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from typing_extensions import override

from .faunadb.objects import FaunaTime, Native, Query, Ref, SetRef
from .faunadb.query import Expr
from .typedefs import Document

T = TypeVar("T")

FaunaKey = Literal[
    "@ref", "@obj", "@set", "@query", "@ts", "@date", "@bytes", "@index", "@class"
]


def _parse_json_hook(dct: Dict[FaunaKey, Any]):
    if "@ref" in dct:
        ref = dct["@ref"]
        if not "collection" in ref and not "database" in ref:
            return Native.from_name(ref["id"])
        return Ref(ref["id"], ref.get("collection"), ref.get("database"))
    if "@obj" in dct:
        return dct["@obj"]
    if "@set" in dct:
        return SetRef(dct["@set"])
    if "@query" in dct:
        return Query(dct["@query"])
    if "@ts" in dct:
        return FaunaTime(dct["@ts"])
    if "@date" in dct:
        return parse_date(dct["@date"]).date()
    if "@bytes" in dct:
        return bytearray(urlsafe_b64decode(dct["@bytes"].encode()))
    return dct


def parse_json(json_string):
    try:
        return loads(json_string, object_hook=_parse_json_hook)
    except ValueError:
        pass


def to_json(dct, pretty=True, sort_keys=True):
    if pretty:
        return dumps(
            dct,
            cls=FaunaJSONEncoder,
            sort_keys=True,
            indent=4,
            separators=(", ", ": "),
            allow_nan=True,
            ensure_ascii=True,
        )
    return dumps(
        dct,
        cls=FaunaJSONEncoder,
        sort_keys=sort_keys,
        separators=(",", ":"),
        allow_nan=False,
        ensure_ascii=True,
        exclude_none=True,
    )


class FaunaJSONEncoder(JSONEncoder):
    @override
    def default(self, obj):
        if isinstance(obj, (Ref, SetRef, FaunaTime, Query)):
            return obj.to_fauna_json()
        if isinstance(obj, Expr):
            return obj.to_fauna_json()
        elif isinstance(obj, datetime):
            return obj.astimezone().isoformat()
        elif isinstance(obj, date):
            return {"@date": obj.isoformat()}
        elif isinstance(obj, (bytes, bytearray)):
            _val = None
            try:
                _val = obj.decode()
            except:
                _val = urlsafe_b64encode(obj).decode()  # pylint: disable=all
            return {"@bytes": _val}
        elif isinstance(obj, BaseModel):
            return obj.dict()
        elif isinstance(obj, Enum):
            return obj.value
        elif isinstance(obj, UUID):
            return {"@uuid": str(obj)}
        else:
            return super().default(obj)


class JSONModel(BaseModel):
    def to_dict(self, **kwargs):
        return parse_json(self.to_json(**kwargs))

    def to_json(self, **kwargs) -> str:
        return to_json(super().dict(exclude_none=True, **kwargs))

    @override
    def dict(self, **kwargs):
        return self.to_dict(**kwargs)

    @override
    def json(self, **kwargs) -> str:
        return self.to_json(exclude_none=True, **kwargs)


def jsonable_encoder(
    obj: Any,
    *,
    include: List[str] = [],
    exclude: List[str] = [],
    by_alias: bool = False,
    skip_defaults: bool = False,
    custom_encoder: Any = None,
) -> Any:
    """
    Convert any object to a JSON-serializable object.

    This function is used by Aiofauna to convert objects to JSON-serializable objects.

    It supports all the types supported by the standard json library, plus:

    * datetime.datetime
    * datetime.date
    * datetime.time
    * uuid.UUID
    * enum.Enum
    * pydantic.BaseModel
    """

    if custom_encoder is None:
        custom_encoder = FaunaJSONEncoder

    if obj is str:
        return "string"
    if obj is int or obj is float:
        return "integer"
    if obj is bool:
        return "boolean"
    if obj is None:
        return "null"
    if obj is list:
        return "array"
    if obj is dict:
        return "object"
    if obj is bytes:
        return "binary"
    if obj is datetime:
        return "date-time"
    if obj is date:
        return "date"
    if obj is time:
        return "time"
    if obj is UUID:
        return "uuid"
    if obj is Enum:
        return "enum"
    if isinstance(obj, (str, int, float, bool, type(None))):
        return obj
    if isinstance(obj, (list, tuple, set, frozenset)):
        return [
            jsonable_encoder(
                v,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                custom_encoder=custom_encoder,
            )
            for v in obj
        ]
    if isinstance(obj, dict):
        return {
            jsonable_encoder(
                k,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                custom_encoder=custom_encoder,
            ): jsonable_encoder(
                v,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                custom_encoder=custom_encoder,
            )
            for k, v in obj.items()
        }
    if isinstance(obj, bytes):
        return base64.b64encode(obj).decode()
    if isinstance(obj, (set, frozenset)):
        return [
            jsonable_encoder(
                v,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                custom_encoder=custom_encoder,
            )
            for v in obj
        ]
    if isinstance(obj, datetime):
        return obj.isoformat()
    if isinstance(obj, Enum):
        return obj.value
    if isinstance(obj, UUID):
        return str(obj)
    if isinstance(obj, type):
        return jsonable_encoder(
            obj.__name__,
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            custom_encoder=custom_encoder,
        )

    return custom_encoder().default(obj)



"""




AioFauna




"""


__version__ = (0, 2, 0)


__author__ = "obahamonde"


__license__ = "MIT"


__doc__ = """
---
title: AioFauna
---

# AioFauna

🚀 Introducing aiofauna: A full-stack framework built on top of Aiohttp, Pydantic and FaunaDB.

🔥 Inspired by FastAPI focuses on Developer Experience, Productivity and Versatility.

🌟 Features:

✅ Supports Python 3.7+, comes with an opinionated ODM (Object Document Mapper) out of the box for FaunaDB that abstracts out complex FQL expressions into pythonic, fully typed asynchronous methods for all CRUD operations.

✅ Powerful and Scalable: Being built on top of Aiohttp an asyncio based http server/client and FaunaDB an scalable serverless database for modern applications allows for powerful and seamless integrations.

✅ Performant: As a framework built on top of Aiohttp it leverages the power of asyncio and the fastest python `APIClient` built on top of aiohttp with Lazy Loading and session sharing, plus the ubiquiness of FaunaDB to achieve high performance.

✅ Automatic Swagger UI generation: Automatic generation of interactive Swagger UI documentation for instant testing of your `APIServer`, exposed at the `/docs` path.

✅ SSE (Server Sent Events): Built-in support for SSE (Server Sent Events) for real-time streaming of data from FaunaDB to your application, syntactic sugar through the `@sse` decorator.

✅ Websockets: Built-in support for Websockets for real-time bidirectional communication between your application and the resources served by AioFauna `APIServer`, syntactic sugar through the `@websocket` decorator.

✅ Robust data validation: Utilizes the rich features of Pydantic for data validation and serialization.

✅ Auto-provisioning: Automatic management of indexes, unique indexes, and collections with `FaunaModel` ODM.

✅ Full JSON communication: Focus on your data, don't worry about the communication protocol. Your `APIServer` will receive and return JSON.

✅ APIRouter: AioFauna comes with a powerful APIRouter class that allows you to organize your `APIServer` endpoints with path operations, path parameters, query parameters, request body, status codes, and decorated view functions.

✅ Inspired by fastapi, you will work with almost the same syntax and features like path operations, path parameters, query parameters, request body, status codes, `/docs` automatic interactive API documentation, and decorated view functions and automatic serialization and deserialization of your data.

💡 With aiofauna, you can build fast, scalable, and reliable modern applications, avoiding decision fatigue and focusing on what really matters, your data and your business logic.

📚 Check out the aiofauna library, and start building your next-gen applications today! 🚀

#Python #FaunaDB #Async #Pydantic #aiofauna

⚙️ If you are using a synchronous framework check out [Faudantic](https://github.com/obahamonde/faudantic) for a similar experience with FaunaDB and Pydantic.

📦 [PyPi](https://pypi.org/project/aiofauna/)

📦 [Demo](https://natal-ia.netlify.app) (LLM smart agent)

📦 [GitHub](https://github.com/obahamonde/aiofauna)

"""


from typing import *  # pylint: disable=wildcard-import,unused-wildcard-import

from aiohttp.web import WebSocketResponse
from aiohttp_sse import EventSourceResponse
from typing_extensions import ParamSpec

from .client import APIClient, FaunaClient
from .docs import FileField
from .faunadb import query as q
from .helpers import async_cpu, async_io, asyncio, template, component
from .json import FaunaJSONEncoder, parse_json, to_json
from .odm import FaunaModel
from .server import APIRouter, APIServer, Request, Response
from .typedefs import Document, Field, LazyProxy, Component
from .utils import chunker, handle_errors, process_time, setup_logging
from .markdown import md_component
from .asgi import ASGIServer


"""REST API Module with automatic OpenAPI generation."""
import asyncio
from inspect import signature
from typing import Callable, List

from aiohttp.typedefs import Handler
from aiohttp.web import Request, RouteDef, RouteTableDef

from .docs import extract, load, transform
from .helpers import do_response


class APIRouter(RouteTableDef):
    """Aiohttp Application with automatic OpenAPI generation."""

    def __init__(self, *args, prefix: str = "", tags: List[str] = [], **kwargs):
        super().__init__(*args, **kwargs)
        self.prefix = prefix
        self.tags = tags
        self.openapi = {
            "paths": {},
            "components": {"schemas": {}},
            "tags": [{"name": tag} for tag in tags if tag],
        }
        self._route_open_api_params = {}
        self.routes: List[RouteDef] = []

    def document(self, path: str, method: str):
        """

        Decorator to document a function.

        """

        def decorator(func: Handler):
            sig = signature(func)
            params = sig.parameters
            open_api_params = extract(params.copy(), path)
            self._route_open_api_params[(path, method)] = open_api_params
            transform(self.openapi, path, method, func, open_api_params)

            async def wrapper(*args, **kwargs):
                request: Request = args[0]
                args = args[1:]
                args_to_apply = await load(request, params.copy())
                definitive_args = {}
                for name, param in params.items():
                    if name in args_to_apply:
                        definitive_args[name] = args_to_apply[name]
                    elif param.default is not param.empty:
                        definitive_args[name] = param.default
                    else:
                        raise ValueError(
                            f"Missing parameter {name} for {func.__name__}"
                        )
                if asyncio.iscoroutinefunction(func):
                    response = await func(*args, **kwargs, **definitive_args)
                else:
                    response = func(*args, **kwargs, **definitive_args)
                return do_response(response)

            func.injectable = True
            wrapper._handler = func
            return wrapper

        return decorator

    def route(self, path: str, method: str):
        """Decorator to register a new route."""

        def decorator(func: Handler, **kwargs):
            route = self.document(self.prefix + path, method)(func)
            self.routes.append(RouteDef(method, path, route, kwargs))
            return route

        return decorator

    def get(self, path: str, **kwargs) -> Callable[[Handler], Handler]:
        """GET decorator"""

        def decorator(func: Handler):
            self.route(path, "GET", **kwargs)(func)
            return func

        return decorator

    def post(self, path: str, **kwargs) -> Callable[[Handler], Handler]:
        """POST decorator"""

        def decorator(func: Handler):
            self.route(path, "POST", **kwargs)(func)
            return func

        return decorator

    def put(self, path: str, **kwargs) -> Callable[[Handler], Handler]:
        """PUT decorator"""

        def decorator(func: Handler):
            self.route(path, "PUT", **kwargs)(func)
            return func

        return decorator

    def delete(self, path: str, **kwargs) -> Callable[[Handler], Handler]:
        """DELETE decorator"""

        def decorator(func: Handler):
            self.route(path, "DELETE", **kwargs)(func)
            return func

        return decorator

    def patch(self, path: str, **kwargs) -> Callable[[Handler], Handler]:
        """PATCH decorator"""

        def decorator(func: Handler):
            self.route(path, "PATCH", **kwargs)(func)
            return func

        return decorator

    def head(self, path: str, **kwargs) -> Callable[[Handler], Handler]:
        """HEAD decorator"""

        def decorator(func: Handler):
            self.route(path, "HEAD", **kwargs)(func)
            return func

        return decorator

    def options(self, path: str, **kwargs) -> Callable[[Handler], Handler]:
        """OPTIONS decorator"""

        def decorator(func: Handler):
            self.route(path, "OPTIONS", **kwargs)(func)
            return func

        return decorator



import os
import subprocess
import logging
import click
import shutil
from rich.console import Console
from rich.logging import RichHandler
from rich.pretty import install
from rich.traceback import install as ins


def setup_logging(name: str) -> logging.Logger:
    """
    Set's up logging using the Rich library for pretty and informative terminal logs.

    Arguments:
    name -- Name for the logger instance. It's best practice to use the name of the module where logger is defined.
    """
    install()
    ins()
    console = Console(record=True, force_terminal=True)
    console_handler = RichHandler(
        console=console,
        show_time=True,
        show_path=True,
        markup=True,
        rich_tracebacks=True,
        tracebacks_show_locals=True,
        tracebacks_extra_lines=2,
        tracebacks_theme="monokai",
        show_level=False,
    )
    console_handler.setFormatter(logging.Formatter("%(message)s"))
    console_handler.setLevel(logging.INFO)
    logging.basicConfig(level=logging.INFO, handlers=[console_handler])
    logger_ = logging.getLogger(name)
    logger_.setLevel(logging.INFO)
    return logger_


logger = setup_logging(__name__)
print = logger.info


@click.group()
def main():
    pass


def print_tree(dirname, pref=""):
    for item in os.listdir(dirname):
        if item.startswith("."):
            continue
        if item == os.listdir(dirname)[-1]:
            print(f"{pref}└── {item}")
            if os.path.isdir(os.path.join(dirname, item)):
                print_tree(os.path.join(dirname, item), pref + "    ")
        else:
            print(f"{pref}├── {item}")
            if os.path.isdir(os.path.join(dirname, item)):
                print_tree(os.path.join(dirname, item), pref + "│   ")


@main.command()
def tree():
    """Prints the directory tree of the current working directory."""
    print_tree(os.getcwd())


@main.command()
def rmcache():
    """Removes all __pycache__ directories in the current working directory and its subdirectories."""
    for root, dirs, files in os.walk(os.getcwd()):
        if "__pycache__" in dirs:
            pycache_path = os.path.join(root, "__pycache__")
            print(f"Removing {pycache_path}")
            shutil.rmtree(pycache_path)


@main.command()
@click.option(
    "--port",
    default=8080,
    type=int,
    help="Port in which the application will listen for requests.",
)
@click.option(
    "--host",
    default="0.0.0.0",
    type=str,
    help="Host that will run the application process.",
)
@click.option(
    "--reload",
    default=True,
    flag_value=True,
    help="When enabled, the server will reload on change.",
)
@click.option(
    "--worker",
    default=1,
    type=int,
    help="Number of workers to spawn. This is the number of processes that will be spawned.",
)
@click.option(
    "--threads",
    default=1,
    type=int,
    help="Number of threads to spawn. This is the number of threads that will be spawned per worker.",
)
@click.argument("name", default="main:app", type=str, nargs=1, required=False)
def run(name, port, host, reload, worker, threads):
    """Spins up a Gunicorn server which serves the app.\n
    It comes with the following options:\n
     --port: Port to run the server on\n
     --host: Host to run the server on\n
     --reload: Reload the server on change\n
     --worker: Number of workers to spawn\n
     --threads: Number of threads to utilize\n\n

     name: Name of the app to run it comes in the format of <module>:<app>, where module is the name of your entry point file without the extension and app is the name of the global variable which contains the `APIServer` instance.\n\n

     Example: aiofauna run main:app --port 5000 --host 0.0.0.0 --reload --worker 1 --threads 1 (This is the default configuration BTW, when you just execute `aiofauna run`)
    """
    subprocess.run(
        [
            "gunicorn",
            "-b",
            f"{host}:{port}",
            "-k",
            "aiohttp.worker.GunicornWebWorker",
            "--reload" if reload else "",
            "--workers",
            str(worker),
            "--threads",
            str(threads),
            name,
        ],
        check=True,
    )


if __name__ == "__main__":
    main()



"""Markdown component decorator for Jinja2 templates."""
import jinja2
import typing_extensions
import typing
from typing import Any, Callable, Coroutine
import functools
from .utils import render_markdown

T = typing.TypeVar("T")
P = typing_extensions.ParamSpec("P")


def md_component(dir: str = "docs", file: str = "base.j2"):
    def decorator(
        func: Callable[P, Coroutine[Any, Any, T]]
    ) -> Callable[P, Coroutine[Any, Any, T]]:
        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(dir),
            autoescape=jinja2.select_autoescape(["html", "xml"]),
            enable_async=True,
        )
        md = render_markdown(func.__doc__)
        template_str = env.from_string(
            """{% extends """
            + "'"
            + file
            + "'"
            + """ %} {% block content %}"""
            + md
            + """{% endblock %}"""
        )

        @functools.wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            return await template_str.render_async(**kwargs)

        return wrapper

    return decorator



"""
Helper functions for aiofauna.
"""
import asyncio
import functools
import typing
from concurrent.futures import ProcessPoolExecutor
from functools import singledispatch
from typing import Any, List, Union, Callable, Awaitable, TypeVar, Generic

from aiohttp.web import Response, json_response
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from typing_extensions import ParamSpec
from jinja2 import Environment, FileSystemLoader, Template, select_autoescape
from .typedefs import Component
from .json import to_json
from .odm import FaunaModel


T = typing.TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")


def template(
    dir: str = "templates", template: str = "base.j2"
) -> Callable[[Callable[P, Awaitable[Response]]], Component]:
    def decorator(func: Callable[P, Awaitable[Response]]):
        env = Environment(
            loader=FileSystemLoader(dir),
            autoescape=select_autoescape(["html", "xml"]),
            enable_async=True,
        )
        doc_template = (
            '{% extends "'
            + template
            + '" %}{% block content %}'
            + func.__doc__
            + "{% endblock %}"
        )

        @functools.wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> Response:
            template = env.from_string(doc_template)
            rendered_template = await template.render_async(await func(*args, **kwargs))
            return Response(text=rendered_template, content_type="text/html")

        return wrapper

    return decorator


def component(func: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[Response]]:
    """
    Decorator to convert a function to a component.
    """

    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> Response:
        context = await func(*args, **kwargs)
        template = Template(func.__doc__)
        html = template.render(context)
        return Response(text=html, content_type="text/html")

    return wrapper


def async_io(
    func: typing.Callable[P, T]
) -> typing.Callable[P, typing.Coroutine[T, Any, Any]]:
    """
    Decorator to convert an IO bound function to a coroutine by running it in a thread pool.
    """

    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        return await asyncio.to_thread(func, *args, **kwargs)

    return wrapper


def async_cpu(
    func: typing.Callable[P, T]
) -> typing.Callable[P, typing.Coroutine[T, Any, Any]]:
    """
    Decorator to convert a CPU bound function to a coroutine by running it in a process pool.
    """

    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        with ProcessPoolExecutor() as pool:
            try:
                return await asyncio.get_running_loop().run_in_executor(
                    pool, func, *args, **kwargs
                )
            except RuntimeError:
                return await asyncio.get_event_loop().run_in_executor(
                    pool, func, *args, **kwargs
                )

    return wrapper


@singledispatch
def do_response(response: T) -> Response:
    """
    Process the response from a view function and return an aiohttp.web.Response object.
    """

    return response


@do_response.register(BaseModel)
def _(response: BaseModel) -> Response:
    return json_response(response.dict(exclude_none=True), dumps=to_json)


@do_response.register(FaunaModel)
def _(response: FaunaModel) -> Response:
    return json_response(response.dict(), dumps=to_json)


@do_response.register(dict)
def _(response: dict) -> Response:
    return json_response(response, dumps=to_json)


@do_response.register(str)
def _(response: str) -> Response:
    if response.startswith("<") and response.endswith(">"):
        return Response(status=200, text=response, content_type="text/html")
    return Response(status=200, text=response, content_type="text/plain")


@do_response.register(bytes)
def _(response: bytes) -> Response:
    return Response(
        status=200, body=response, content_type="application/octet-sse_stream"
    )


@do_response.register(int)
def _(response: int) -> Response:
    return Response(status=200, text=str(response), content_type="text/plain")


@do_response.register(float)
def _(response: float) -> Response:
    return Response(status=200, text=str(response), content_type="text/plain")


@do_response.register(bool)
def _(response: bool) -> Response:
    return Response(status=200, text=str(response), content_type="text/plain")


@do_response.register(list)
def _(response: List[Union[FaunaModel, BaseModel, dict, str, int, float]]) -> Response:
    return json_response(response, dumps=to_json)


@do_response.register(tuple)
def _(response: tuple) -> Response:
    return do_response(list(response))


@do_response.register(set)
def _(response: set) -> Response:
    return do_response(list(response))


@do_response.register(frozenset)
def _(response: frozenset) -> Response:
    return do_response(list(response))


@do_response.register(type(None))
def _(response: None) -> Response:
    return Response(status=200, text="", content_type="text/plain")


@do_response.register(Response)
def _(response: Response) -> Response:
    return response



from __future__ import annotations

import asyncio
from typing import Any, List, Type, TypeVar

from dotenv import load_dotenv
from pydantic import Field

from .client import FaunaClient
from .faunadb import query as q
from .faunadb.errors import FaunaException
from .json import JSONModel
from .utils import gen_emptystr, setup_logging

load_dotenv()

T = TypeVar("T", bound="FaunaModel")


class FaunaModel(JSONModel):
    ref: str = Field(default_factory=gen_emptystr)
    ts: str = Field(default_factory=gen_emptystr)

    class Metadata:
        subclasses: List[Type[FaunaModel]] = []

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls.__doc__ is None:
            cls.__doc__ = f"```json\n{cls.schema_json(indent=2)}\n```"
        cls.Metadata.subclasses.append(cls)

    @classmethod
    @property
    def logger(cls):
        return setup_logging(cls.__name__)

    @classmethod
    async def create_all(cls):
        await asyncio.gather(*[model.provision() for model in cls.Metadata.subclasses])

    @classmethod
    def client(cls):
        return FaunaClient()

    @classmethod
    def q(cls):
        return cls.client().query

    @classmethod
    async def provision(cls) -> bool:
        _q = cls.q()

        try:
            if not await _q(q.exists(q.collection(cls.__name__.lower()))):
                await _q(q.create_collection({"name": cls.__name__.lower()}))

                cls.logger.info("Created collection %s", cls.__name__.lower())

            if not await _q(q.exists(q.index(cls.__name__.lower()))):
                await _q(
                    q.create_index(
                        {
                            "name": cls.__name__.lower(),
                            "source": q.collection(cls.__name__.lower()),
                        }
                    )
                )

                cls.logger.info("Created index %s", cls.__name__.lower())

            for field in cls.__fields__.values():
                if field.field_info.extra.get("unique"):
                    if not await _q(
                        q.exists(q.index(f"{cls.__name__.lower()}_{field.name}_unique"))
                    ):
                        await _q(
                            q.create_index(
                                {
                                    "name": f"{cls.__name__.lower()}_{field.name}_unique",
                                    "source": q.collection(cls.__name__.lower()),
                                    "terms": [{"field": ["data", field.name]}],
                                    "unique": True,
                                }
                            )
                        )

                        cls.logger.info(
                            "Created unique index %s_%s",
                            cls.__name__.lower(),
                            field.name,
                        )
                    continue

                if field.field_info.extra.get("index"):
                    if not await _q(
                        q.exists(q.index(f"{cls.__name__.lower()}_{field.name}"))
                    ):
                        await _q(
                            q.create_index(
                                {
                                    "name": f"{cls.__name__.lower()}_{field.name}",
                                    "source": q.collection(cls.__name__.lower()),
                                    "terms": [{"field": ["data", field.name]}],
                                }
                            )
                        )

                        cls.logger.info(
                            "Created index %s_%s", cls.__name__.lower(), field.name
                        )
                        continue

            return True

        except (FaunaException, KeyError, TypeError, ValueError, Exception) as exc:
            cls.logger.error(exc.__class__.__name__)
            cls.logger.error(exc)

            return False

    @classmethod
    async def find_unique(cls: Type[T], **kwargs: Any) -> T:
        try:
            field, value = list(kwargs.items())[0]
            data = await cls.q()(
                q.get(q.match(q.index(f"{cls.__name__.lower()}_{field}_unique"), value))
            )
            return cls(
                **{
                    **data["data"],  # type: ignore
                    "ref": data["ref"]["@ref"]["id"],  # type: ignore
                    "ts": data["ts"] / 1000,  # type: ignore
                }
            )

        except (FaunaException, KeyError, TypeError, ValueError, Exception) as exc:
            cls.logger.error(exc.__class__.__name__)
            cls.logger.error(exc)
            return None  # type: ignore

    @classmethod
    async def find_many(cls: Type[T], limit: int = 50, **kwargs: Any) -> List[T]:
        try:
            _q = cls.q()
            field, value = list(kwargs.items())[0]
            refs = await _q(
                q.paginate(q.match(q.index(f"{cls.__name__.lower()}_{field}"), value))
            )
            refs = refs["data"][:limit]
            return await asyncio.gather(*[cls.get(item["@ref"]["id"]) for item in refs])

        except (FaunaException, KeyError, TypeError, ValueError, Exception) as exc:
            cls.logger.error(exc.__class__.__name__)
            cls.logger.error(exc)
            return []

    @classmethod
    async def get(cls: Type[T], ref: str) -> T:
        try:
            data = await cls.q()(q.get(q.ref(q.collection(cls.__name__.lower()), ref)))
            return cls(
                **{
                    **data["data"],  # type: ignore
                    "ref": data["ref"]["@ref"]["id"],  # type: ignore
                    "ts": data["ts"] / 1000,  # type: ignore
                }
            )

        except (FaunaException, KeyError, TypeError, ValueError, Exception) as exc:
            cls.logger.error(exc.__class__.__name__)
            cls.logger.error(exc)
            raise exc

    @classmethod
    async def all(cls: Type[T], limit: int = 100, offset: int = 0) -> List[T]:
        try:
            _q = cls.q()

            query = q.paginate(q.match(q.index(f"{cls.__name__.lower()}")))

            refs = (await _q(query))["data"][offset:limit]

            return await asyncio.gather(*[cls.get(item["@ref"]["id"]) for item in refs])

        except (FaunaException, KeyError, TypeError, ValueError, Exception) as exc:
            cls.logger.error(exc.__class__.__name__)
            cls.logger.error(exc)
            raise exc

    @classmethod
    async def delete_one(cls: Type[T], **kwargs: Any) -> bool:
        try:
            field, value = list(kwargs.items())[0]
            _q = cls.q()

            ref = await _q(
                q.get(q.match(q.index(f"{cls.__name__.lower()}_{field}_unique"), value))
            )

            await _q(q.delete(ref))

            return True

        except (FaunaException, KeyError, TypeError, ValueError, Exception) as exc:
            cls.logger.error(exc.__class__.__name__)
            cls.logger.error(exc)
            return False

    @classmethod
    async def delete(cls, ref: str) -> bool:
        try:
            await cls.q()(q.delete(q.ref(q.collection(cls.__name__.lower()), ref)))

            return True

        except (FaunaException, KeyError, TypeError, ValueError, Exception) as exc:
            cls.logger.error(exc.__class__.__name__)
            cls.logger.error(exc)
            return False

    async def create(self: T) -> T:
        try:
            for field in self.__fields__.values():
                if field.field_info.extra.get("unique"):
                    instance = await self.find_unique(
                        **{field.name: self.__dict__[field.name]}
                    )

                    if instance is None:
                        continue

                    if issubclass(instance.__class__, FaunaModel):
                        return instance

            data = await self.__class__.q()(
                q.create(
                    q.collection(self.__class__.__name__.lower()), {"data": self.dict()}
                )
            )

            self.ref = data["ref"]["@ref"]["id"]  # type: ignore

            self.ts = data["ts"] / 1000  # type: ignore
            return self

        except (FaunaException, KeyError, TypeError, ValueError, Exception) as exc:
            self.logger.error(exc.__class__.__name__)
            self.logger.error(exc)
            raise exc

    @classmethod
    async def update(cls: Type[T], ref: str, **kwargs: Any) -> T:
        try:
            instance = await cls.q()(
                q.update(
                    q.ref(q.collection(cls.__name__.lower()), ref),
                    {"data": kwargs.get("kwargs", kwargs)},
                )
            )
            return cls(
                **{
                    **instance["data"],  # type: ignore
                    "ref": instance["ref"]["@ref"]["id"],  # type: ignore
                    "ts": instance["ts"] / 1000,  # type: ignore
                }
            )
        except (FaunaException, KeyError, TypeError) as exc:
            cls.logger.error(exc.__class__.__name__)
            cls.logger.error(exc)
            raise ValueError(f"Field {ref} not found")  # pylint: disable=all

    async def save(self: T) -> T:
        if isinstance(self.ref, str) and len(self.ref) == 18:
            return await self.update(self.ref, kwargs=self.dict())
        return await self.create()

    @classmethod
    async def cleanup(cls):
        await cls.client().cleanup()



import asyncio
from typing import Any, Awaitable, Callable, Literal, MutableMapping, Tuple

from aiohttp import web
from aiohttp.http_parser import RawRequestMessage
from aiohttp.streams import StreamReader
from aiohttp.payload import StringPayload
from multidict import CIMultiDict, CIMultiDictProxy
from yarl import URL

from .server import APIServer

Scope = MutableMapping[str, Any]
Message = MutableMapping[str, Any]
Receive = Callable[[], Awaitable[Message]]
Send = Callable[[Message], Awaitable[None]]

ASGIApp = Callable[[Scope, Receive, Send], Awaitable[None]]


class ASGIServer(APIServer):
    """
    An ASGI-compatible API.

    Currently only supports HTTP and is compatible with ASGI 2.0.

    Tested with:
    `uvicorn`
    `mangum`
    """

    async def asgi(self, scope: Scope, _, send: Send) -> None:
        """ASGI Wrapper."""

        if scope["type"] != "http":
            raise RuntimeError("Only HTTP is supported")

        raw_headers: Tuple[Tuple[bytes, bytes], ...] = scope["headers"]
        headers = CIMultiDict()
        for k, v in raw_headers:
            headers[k.decode().lower()] = v.decode()

        message = RawRequestMessage(
            method=scope["method"],
            path=scope["path"],
            version=scope["http_version"],
            headers=CIMultiDictProxy(headers),
            raw_headers=raw_headers,
            should_close=True,
            compression=None,
            upgrade=False,
            chunked=True,
            url=URL(scope["path"]),
        )

        class _tranport:
            scheme = scope["scheme"]

            def get_extra_info(
                self, info: Literal["sslcontext"]
            ):  # pylint: disable=unused-argument
                """Mock method to match `aiohttp.web.BaseRequestTransport` contract."""
                return None

        class _proto(web.RequestHandler):
            http_version = scope["http_version"]
            transport = _tranport()

        request = web.Request(
            message=message,
            payload=StreamReader(protocol=_proto,limit=2**20),
            protocol=_proto,
            payload_writer=None,
            task=None,
            loop=asyncio.get_running_loop(),
        )

        response = await self._handle(request)

        body = response._body
        if isinstance(body, StringPayload):
            body = body._value
        if isinstance(body, str):
            body = body.encode()

        await send(
            {
                "type": "http.response.start",
                "status": response.status,
                "headers": [
                    (k.encode(), v.encode()) for k, v in response.headers.items()
                ],
            }
        )

        await send({"type": "http.response.body", "body": body, "more_body": False})

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """ASGI callable."""
        return await self.asgi(scope, receive, send)


"""
Exceptions for aiofauna
"""

from json import JSONDecodeError

from aiohttp import (
    ClientConnectionError,
    ClientConnectorError,
    ClientConnectorSSLError,
    ClientError,
    ContentTypeError,
    ServerTimeoutError,
    WSServerHandshakeError,
)
from aiohttp.web_exceptions import HTTPException

from .faunadb.errors import FaunaError, FaunaException


class AiofaunaException(Exception):
    """
    Base exception.
    """

    def __init__(self, message: str = None, *args: object) -> None:
        if message is None:
            message = self.__doc__
        super().__init__(message, *args)


EXCEPTIONS = (
    # aiohttp exceptions
    ClientConnectionError,
    ClientConnectorError,
    ClientConnectorSSLError,
    ClientError,
    ContentTypeError,
    ServerTimeoutError,
    WSServerHandshakeError,
    # aiofauna exceptions
    AiofaunaException,
    # faunadb exceptions
    FaunaError,
    FaunaException,
    # json exceptions
    JSONDecodeError,
    # aiohttp.web_exceptions exceptions
    HTTPException,
    ValueError,
    KeyError,
    TypeError,
    Exception,
    UnicodeError,
    RuntimeError,
)



from __future__ import annotations

import asyncio
import json
import logging
import os
import typing
from dataclasses import dataclass, field
from aiohttp import ClientResponse
from functools import wraps
from threading import Lock
from typing import Any, AsyncGenerator, Dict, List, Literal, Optional, Type, Union

from aiohttp import ClientSession, ClientTimeout, TCPConnector
from aiohttp.client_exceptions import ClientConnectionError, ClientConnectorSSLError
from aiohttp.web_exceptions import HTTPException
from dotenv import load_dotenv
from multidict import CIMultiDict

from .faunadb.errors import FaunaException
from .faunadb.objects import Expr
from .json import to_json
from .typedefs import LazyProxy
from .utils import setup_logging

load_dotenv()

Method = Literal["GET", "POST", "PUT", "PATCH", "DELETE"]
Json = Union[Dict[str, Any], List[Dict[str, Any]]]
MaybeJson = Optional[Json]
Headers = Dict[str, str]
MaybeHeaders = Optional[Headers]


class MissingEnvironmentVariable(Exception):
    """Exception for missing environment variable."""

    ...


if "FAUNA_SECRET" not in os.environ:
    raise MissingEnvironmentVariable(
        "The FAUNA_SECRET environment variable is not set."
    )


FAUNA_SECRET = os.environ["FAUNA_SECRET"]
HEADERS = {
    "Authorization": f"Bearer {FAUNA_SECRET}",
    "Content-type": "application/json",
    "Accept": "application/json",
    "User-Agent": "aiofauna-framework",
}

T = typing.TypeVar("T")

logger = setup_logging(__name__)

FAUNA_EXCEPTIONS = (
    HTTPException,
    FaunaException,
    ValueError,
    KeyError,
    TypeError,
    Exception,
    UnicodeError,
    json.JSONDecodeError,
    RuntimeError,
    ClientConnectionError,
    ClientConnectorSSLError,
)

HTTP_EXCEPTIONS = (
    HTTPException,
    FaunaException,
    ValueError,
    KeyError,
    TypeError,
    Exception,
)


def singleton(cls: typing.Type[T]) -> typing.Type[T]:
    instance = None
    lock = Lock()

    @wraps(cls)
    def wrapper(*args, **kwargs):
        nonlocal instance
        if instance is None:
            with lock:
                if instance is None:
                    instance = cls(*args, **kwargs)
        return instance

    return wrapper


class APIException(HTTPException):
    """Base class for all exceptions raised by an API client."""

    def __init__(self, message: str, status_code: int = 500) -> None:
        self.message = message
        self.status_code = status_code
        super().__init__(
            text=json.dumps({"message": message, "status": status_code}),
            content_type="application/json",
        )


@dataclass(init=True, repr=True, unsafe_hash=False, frozen=False)
class ConnectorConfig:
    ssl: bool = field(default=False)
    limit: int = field(default=2048)
    keepalive_timeout: int = field(default=60)


@dataclass(init=True, repr=True, unsafe_hash=False, frozen=False)
class APIConfig:
    base_url: str
    headers: Headers
    logger: logging.Logger = field(default=setup_logging(__name__))
    exception_class: Type[HTTPException] = field(default=APIException)
    session: Optional[ClientSession] = field(default=None)
    connector_config: ConnectorConfig = field(default_factory=ConnectorConfig)
    read_bufsize: int = field(default=2**16)
    connector_owner: bool = field(default=True)
    trust_env: bool = field(default=True)


@dataclass(init=True, repr=True, unsafe_hash=False, frozen=False)
class FaunaClient(LazyProxy[ClientSession]):
    """
    FaunaDB Client

    Args:
        config (APIConfig): APIConfig object
    """

    config: APIConfig = APIConfig(
        base_url="https://db.fauna.com",
        headers=HEADERS,
        logger=setup_logging(__name__),
        exception_class=APIException,
        session=None,
        connector_config=ConnectorConfig(),
    )
    session_creation_lock = asyncio.Lock()

    def __load__(self) -> ClientSession:
        if self.config.session is None:
            return ClientSession(
                "https://db.fauna.com",
                headers=CIMultiDict(self.config.headers),  # type: ignore
                connector=TCPConnector(
                    keepalive_timeout=self.config.connector_config.keepalive_timeout,
                    ssl=self.config.connector_config.ssl,
                    limit=self.config.connector_config.limit,
                ),
                connector_owner=self.config.connector_owner,
                trust_env=self.config.trust_env,
                read_bufsize=self.config.read_bufsize,
            )
        return self.config.session

    async def query(self, expr: Expr) -> Any:
        async with self.session_creation_lock:
            if self.config.session is None:
                self.config.session = self.__load__()
        session = self.config.session
        async with session.post(
            "/",
            data=to_json(expr),
        ) as response:
            try:
                data = await response.json()
                self.config.logger.info(data)
                if data.get("resource") is not None:
                    return data["resource"]
                if data.get("error") is not None:
                    return data["error"]
                return data

            except FAUNA_EXCEPTIONS as exc:  # pylint:disable=all
                self.config.logger.error(exc)
                raise self.config.exception_class from exc

    async def stream(self, expr: Expr) -> AsyncGenerator[str, None]:
        session = self.__load__()
        async with session.post(
            "",
            data=to_json(expr),
            headers={
                "Authorization": f"Bearer {self.secret}",
                "Content-type": "application/json",
                "Accept": "text/event-stream",
                "Keep-Alive": "timeout=60, max=900",
                "Connection": "keep-alive",
                "Cache-Control": "no-cache",
                "X-Last-Seen-Txn": "0",
                "X-Request-By": "aiofauna",
                "X-Query-By": "@obahamonde",
            },
        ) as response:
            async for chunk in response.content.iter_any():
                try:
                    yield chunk.decode()
                except HTTP_EXCEPTIONS as exc:
                    self.config.logger.error(exc)
                    yield str(exc)
                    raise self.config.exception_class from exc

    async def cleanup(self):
        if self.config.session is not None:
            await self.config.session.close()


@dataclass(init=True, repr=True, unsafe_hash=False, frozen=False)
class APIClient(LazyProxy[ClientSession]):
    """
    HTTP Client:
    Base class to create HTTP clients that wrap `aiohttp.ClientSession`.
    Provides Lazy Loading through proxy objects and session reuse using the singleton pattern.
    Constructor Signature:
    `base_url` (str): Base URL of the API. For example: `https://api.openai.com`. Must the an absolute URL.
    `headers` (dict): Headers that will proxide the authentication credentials, content_type, etc.
    """

    base_url: str = field(init=True, repr=True)
    headers: Dict[str, str] = field(default_factory=dict)
    _subclasses: Optional[List[Type[APIClient]]] = None
    _session_creation_lock = asyncio.Lock()
    _session: Optional[ClientSession] = None

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls._subclasses is None:
            cls._subclasses = []
        cls = dataclass(cls, init=True, repr=True, unsafe_hash=False, frozen=False)
        cls._subclasses.append(cls)

    @classmethod
    async def cleanup(cls):
        if cls._subclasses is not None:
            tasks = []
            for subclass in cls._subclasses:
                if subclass._session is not None:
                    tasks.append(subclass._session.close())
            await asyncio.gather(*tasks)

    async def __load__(self) -> ClientSession:
        async with self._session_creation_lock:
            if self._session is None:
                self._session = ClientSession(
                    base_url=self.base_url,
                    headers=CIMultiDict(self.headers),
                    response_class=ClientResponse,
                    connector=TCPConnector(
                        keepalive_timeout=60,
                        ssl=False,
                        limit=2048,
                    ),
                    timeout=ClientTimeout(total=60),
                    connector_owner=False,
                    trust_env=False,
                    read_bufsize=2**18,
                )
        return self._session

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.base_url}>"

    async def fetch(
        self,
        url: str,
        method: Method = "GET",
        json: MaybeJson = None,
    ):
        session = await self.__load__()
        async with session.request(
            method, url, headers=self.headers, json=json
        ) as response:
            try:
                return await response.json()
            except (
                HTTP_EXCEPTIONS
            ) as exc:  # pylint:disable=broad-exception-caught, unused-variable
                logger.error(exc)
                raise APIException(str(exc)) from exc

    async def text(self, url: str, method: Method = "GET", json: MaybeJson = None):
        session = await self.__load__()
        async with session.request(
            method, url, headers=self.headers, json=json
        ) as response:
            try:
                return await response.text()
            except (
                HTTP_EXCEPTIONS
            ) as exc:  # pylint:disable=broad-exception-caught, unused-variable
                logger.error(exc)
                raise APIException(str(exc)) from exc

    async def stream(
        self,
        url: str,
        method: Method = "GET",
        json: MaybeJson = None,
    ):
        session = await self.__load__()
        async with session.request(
            method, url, headers=self.headers, json=json
        ) as response:
            async for chunk in response.content.iter_any():
                try:
                    yield chunk.decode()
                except HTTP_EXCEPTIONS as exc:
                    logger.error(exc)
                    yield str(exc)
                    raise APIException(str(exc)) from exc

    async def get(self, url: str):
        return await self.fetch(url=url, method="GET")

    async def post(self, url: str, json: MaybeJson = None):
        return await self.fetch(url=url, method="POST", json=json)

    async def put(self, url: str, json: MaybeJson = None):
        return await self.fetch(url=url, method="PUT", json=json)

    async def patch(self, url: str, json: MaybeJson = None):
        return await self.fetch(url=url, method="PATCH", json=json)

    async def delete(self, url: str, json: MaybeJson = None):
        return await self.fetch(url=url, method="DELETE", json=json)

    def update_headers(self, headers: Dict[str, str]):
        """
        Update the headers used by this API client.

        Returns:
            self: Allows method chaining.
        """
        self.headers.update(headers)
        return self



"""
Types used in queries and responses.
See the `docs <https://app.fauna.com/documentation/reference/queryapi#simple-type>`__.
"""
from datetime import datetime

from iso8601 import parse_date

from .query import Expr


class Ref(Expr):
    """
    ```python
    Ref(id, cls=None, db=None)
    ```
    A reference to a document in a collection or index.

    :param id: The document's ID.
    :param cls: The collection or index class.
    :param db: The database.

    `Ref`

    Is a special type in FaunaDB. It is used to represent a document in a collection or index.

    It is serialized to JSON as an object with the `@ref` key. Passing the `id` to the response.

    """

    def __init__(self, id, cls=None, db=None):
        if id is None:
            raise ValueError("The Ref must have an id.")

        value = {"id": id}

        if cls != None:
            value["collection"] = cls

        if db != None:
            value["database"] = db

        super(Ref, self).__init__(value)

    def collection(self):
        """
        Gets the collection part out of the Ref.
        """
        return self.value.get("collection")

    def database(self):
        """
        Gets the database part out of the Ref.
        """
        return self.value.get("database")

    def id(self):
        """
        Gets the id part out of the Ref.
        """
        return self.value["id"]

    def to_fauna_json(self):
        return {"@ref": self.value}

    def __str__(self):
        col = (
            f", collection={self.value['collection']}"
            if "collection" in self.value
            else ""
        )
        db = (
            f", database={self.value.get('database')}"
            if "database" in self.value
            else ""
        )
        return "Ref(id=%s%s%s)" % (self.value["id"], col, db)

    def __repr__(self):
        col = (
            f", collection={self.value['collection']}"
            if "collection" in self.value
            else ""
        )
        db = (
            f", database={self.value.get('database')}"
            if "database" in self.value
            else ""
        )
        return "Ref(id=%s%s%s)" % (self.value["id"], col, db)

    def __eq__(self, other):
        return isinstance(other, Ref) and self.value == other.value

    def __ne__(self, other):
        # pylint: disable=unneeded-not
        return not self == other


class Native(object):
    COLLECTIONS = Ref("collections")
    INDEXES = Ref("indexes")
    DATABASES = Ref("databases")
    FUNCTIONS = Ref("functions")
    KEYS = Ref("keys")
    TOKENS = Ref("tokens")
    CREDENTIALS = Ref("credentials")
    ROLES = Ref("roles")
    ACCESS_PROVIDERS = Ref("access_providers")

    def __init__(self):
        raise TypeError

    @classmethod
    def from_name(cls, name):
        return getattr(cls, name.upper(), Ref(name))


class SetRef(Expr):
    """
    FaunaDB Set.
    This represents a set returned as part of a response.
    For query sets see :doc:`query`.
    """

    def __init__(self, set_ref):
        if isinstance(set_ref, Expr):
            value = set_ref.value
        else:
            value = set_ref

        super(SetRef, self).__init__(value)

    def to_fauna_json(self):
        return {"@set": self.value}

    def __repr__(self):
        return f"SetRef({repr(self.value)})"

    def __eq__(self, other):
        return isinstance(other, SetRef) and self.value == other.value

    def __ne__(self, other):
        # pylint: disable=unneeded-not
        return not self == other


class FaunaTime(Expr):
    """
    FaunaDB time. See the `docs <https://app.fauna.com/documentation/reference/queryapi#special-type>`__.

    For dates, regular :class:`datetime.date` objects are used.
    """

    def __init__(self, value):
        """
        :param value:
          If a :class:`datetime.datetime` is passed, it is converted to a string.
          Must include an offset.
        """
        if isinstance(value, datetime):
            if value.utcoffset() is None:
                raise ValueError("FaunaTime requires offset-aware datetimes")
            value = value.isoformat()

        # Convert +00:00 offset to zulu for comparison equality
        # We don't check for +0000 or +00 as they are not valid in FaunaDB
        super(FaunaTime, self).__init__(value.replace("+00:00", "Z"))

    def to_datetime(self):
        """
        Convert to an offset-aware datetime object.
        This is lossy as datetimes have microsecond rather than nanosecond precision.
        """
        return parse_date(self.value)

    def to_fauna_json(self):
        return {"@ts": self.value}

    def __repr__(self):
        return "FaunaTime(%s)" % repr(self.value)

    def __eq__(self, other):
        return isinstance(other, FaunaTime) and self.value == other.value

    def __ne__(self, other):
        # pylint: disable=unneeded-not
        return not self == other


class Query(Expr):
    """
    Represents a `@query` type in FaunaDB.
    See the `docs <https://app.fauna.com/documentation/reference/queryapi#special-type>`__.
    """

    def to_fauna_json(self):
        return {"@query": self.value}

    def __repr__(self):
        return "Query(%s)" % repr(self.value)

    def __eq__(self, other):
        return isinstance(other, Query) and self.value == other.value

    def __ne__(self, other):
        # pylint: disable=unneeded-not
        return not self == other



"""zLogging and error handling utilities for the OpenAI Function Python package."""
import functools
import logging
from time import perf_counter
from typing import Any, Callable, Coroutine, Generator, Sequence, TypeVar, cast
from aiohttp.web_response import Response
from aiohttp.web_exceptions import HTTPException
from rich.console import Console
from rich.logging import RichHandler
from rich.pretty import install
from rich.traceback import install as ins
from typing import Optional
from typing_extensions import ParamSpec
from markdown_it import MarkdownIt
from markdown_it.renderer import RendererHTML
from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexers import get_lexer_by_name
from pygments.styles import get_style_by_name
from jinja2 import Environment, FileSystemLoader, Template
from .exceptions import EXCEPTIONS

T = TypeVar("T")
P = ParamSpec("P")


def setup_logging(name: str) -> logging.Logger:
    """
    Set's up logging using the Rich library for pretty and informative terminal logs.

    Arguments:
    name -- Name for the logger instance. It's best practice to use the name of the module where logger is defined.
    """
    install()
    ins()
    console = Console(record=True, force_terminal=True)
    console_handler = RichHandler(
        console=console,
        show_time=True,
        show_path=True,
        markup=True,
        rich_tracebacks=True,
        tracebacks_show_locals=True,
        tracebacks_extra_lines=2,
        tracebacks_theme="monokai",
        show_level=False,
    )
    console_handler.setFormatter(logging.Formatter("%(message)s"))
    console_handler.setLevel(logging.INFO)
    logging.basicConfig(level=logging.INFO, handlers=[console_handler])
    logger_ = logging.getLogger(name)
    logger_.setLevel(logging.INFO)
    return logger_


logger = setup_logging(__name__)


def process_time(
    func: Callable[P, Coroutine[Any, Any, T]]
) -> Callable[P, Coroutine[Any, Any, T]]:
    """
    A decorator to measure the execution time of a coroutine.

    Arguments:
    func -- The coroutine whose execution time is to be measured.
    """

    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        """
        Wrapper function to time the function call.
        """
        start = perf_counter()
        result = await func(*args, **kwargs)
        end = perf_counter()
        logger.info("Time taken to execute %s: %s seconds", func.__name__, end - start)
        return result

    return wrapper


def handle_errors(
    func: Callable[P, Coroutine[Any, Any, T]]
) -> Callable[P, Coroutine[Any, Any, T]]:
    """
    A decorator to handle errors in a coroutine.

    Arguments:
    func -- The coroutine whose errors are to be handled.
    """

    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        """
        Wrapper function to handle errors in the function call.
        """
        try:
            return await func(*args, **kwargs)
        except EXCEPTIONS as exc:
            logger.error(exc.__class__.__name__)
            logger.error(str(exc))
            raise HTTPException(reason=str(exc)) from exc

    return wrapper


def chunker(seq: Sequence[T], size: int) -> Generator[Sequence[T], None, None]:
    """
    A generator function that chunks a sequence into smaller sequences of the given size.

    Arguments:
    seq -- The sequence to be chunked.
    size -- The size of the chunks.
    """
    return (seq[pos : pos + size] for pos in range(0, len(seq), size))


def gen_emptystr() -> str:
    """
    A generator function that returns an empty string.
    """
    return cast(str, None)


def highlight_code(code, lang, opts, env):
    """
    A function that highlights code using pygments.
    """
    if not lang:
        lang = "md"
    lexer = get_lexer_by_name(lang, stripall=True)
    formatter = HtmlFormatter(**opts)
    return highlight(code, lexer, formatter)


def highlight_code(code: str, lang: str, _options: dict) -> str:
    if not lang:
        lexer = get_lexer_by_name("md")
    lexer = get_lexer_by_name(lang)
    style = get_style_by_name("monokai")
    formatter = HtmlFormatter(style=style)
    return highlight(code, lexer, formatter)


def render_markdown(markdown: str) -> str:
    markdownit = MarkdownIt(
        "js-default",
        {
            "html": True,
            "linkify": True,
            "typographer": True,
            "highlight": highlight_code,
            "renderer": RendererHTML(),
        },
    )
    return markdownit.render(markdown)
